@@ -19,7 +19,7 @@ fn error_from_signal(recipe: &str, line_number: Option<usize>, exit_status: Exit
 /// A recipe, e.g. `foo: bar baz`
 #[derive(PartialEq, Debug, Clone, Serialize)]
 pub(crate) struct Recipe<'src, D = Dependency<'src>> {
-  pub(crate) attributes: BTreeSet<Attribute<'src>>,
+  pub(crate) attributes: AttributeSet<'src>,
   pub(crate) body: Vec<Line<'src>>,
   pub(crate) dependencies: Vec<D>,
   pub(crate) doc: Option<String>,
@@ -66,20 +66,20 @@ impl<'src, D> Recipe<'src, D> {
   }
 
   pub(crate) fn confirm(&self) -> RunResult<'src, bool> {
-    for attribute in &self.attributes {
-      if let Attribute::Confirm(prompt) = attribute {
-        if let Some(prompt) = prompt {
-          eprint!("{} ", prompt.cooked);
-        } else {
-          eprint!("Run recipe `{}`? ", self.name);
-        }
-        let mut line = String::new();
-        std::io::stdin()
-          .read_line(&mut line)
-          .map_err(|io_error| Error::GetConfirmation { io_error })?;
-        let line = line.trim().to_lowercase();
-        return Ok(line == "y" || line == "yes");
+    if let Some(Attribute::Confirm(ref prompt)) =
+      self.attributes.get(AttributeDiscriminant::Confirm)
+    {
+      if let Some(prompt) = prompt {
+        eprint!("{} ", prompt.cooked);
+      } else {
+        eprint!("Run recipe `{}`? ", self.name);
       }
+      let mut line = String::new();
+      std::io::stdin()
+        .read_line(&mut line)
+        .map_err(|io_error| Error::GetConfirmation { io_error })?;
+      let line = line.trim().to_lowercase();
+      return Ok(line == "y" || line == "yes");
     }
     Ok(true)
   }
@@ -97,7 +97,7 @@ impl<'src, D> Recipe<'src, D> {
   }
 
   pub(crate) fn is_public(&self) -> bool {
-    !self.private && !self.attributes.contains(&Attribute::Private)
+    !self.private && !self.attributes.contains(AttributeDiscriminant::Private)
   }
 
   pub(crate) fn is_script(&self) -> bool {
@@ -105,19 +105,22 @@ impl<'src, D> Recipe<'src, D> {
   }
 
   pub(crate) fn takes_positional_arguments(&self, settings: &Settings) -> bool {
-    settings.positional_arguments || self.attributes.contains(&Attribute::PositionalArguments)
+    settings.positional_arguments
+      || self
+        .attributes
+        .contains(AttributeDiscriminant::PositionalArguments)
   }
 
   pub(crate) fn change_directory(&self) -> bool {
-    !self.attributes.contains(&Attribute::NoCd)
+    !self.attributes.contains(AttributeDiscriminant::NoCd)
   }
 
   pub(crate) fn enabled(&self) -> bool {
-    let linux = self.attributes.contains(&Attribute::Linux);
-    let macos = self.attributes.contains(&Attribute::Macos);
-    let openbsd = self.attributes.contains(&Attribute::Openbsd);
-    let unix = self.attributes.contains(&Attribute::Unix);
-    let windows = self.attributes.contains(&Attribute::Windows);
+    let linux = self.attributes.contains(AttributeDiscriminant::Linux);
+    let macos = self.attributes.contains(AttributeDiscriminant::Macos);
+    let openbsd = self.attributes.contains(AttributeDiscriminant::Openbsd);
+    let unix = self.attributes.contains(AttributeDiscriminant::Unix);
+    let windows = self.attributes.contains(AttributeDiscriminant::Windows);
 
     (!windows && !linux && !macos && !openbsd && !unix)
       || (cfg!(target_os = "linux") && (linux || unix))
@@ -128,8 +131,16 @@ impl<'src, D> Recipe<'src, D> {
       || (cfg!(windows) && windows)
   }
 
-  fn print_exit_message(&self) -> bool {
-    !self.attributes.contains(&Attribute::NoExitMessage)
+  fn print_exit_message(&self, settings: &Settings) -> bool {
+    if self.attributes.contains(AttributeDiscriminant::ExitMessage) {
+      true
+    } else if settings.no_exit_message {
+      false
+    } else {
+      !self
+        .attributes
+        .contains(AttributeDiscriminant::NoExitMessage)
+    }
   }
 
   fn working_directory<'a>(&'a self, context: &'a ExecutionContext) -> Option<PathBuf> {
@@ -149,7 +160,7 @@ impl<'src, D> Recipe<'src, D> {
   }
 
   fn no_quiet(&self) -> bool {
-    self.attributes.contains(&Attribute::NoQuiet)
+    self.attributes.contains(AttributeDiscriminant::NoQuiet)
   }
 
   pub(crate) fn run<'run>(
@@ -199,11 +210,11 @@ impl<'src, D> Recipe<'src, D> {
       }
       let mut evaluated = String::new();
       let mut continued = false;
-      let quiet_line = lines.peek().map_or(false, |line| line.is_quiet());
-      let infallible_line = lines.peek().map_or(false, |line| line.is_infallible());
+      let quiet_line = lines.peek().is_some_and(|line| line.is_quiet());
+      let infallible_line = lines.peek().is_some_and(|line| line.is_infallible());
 
       let comment_line = context.module.settings.ignore_comments
-        && lines.peek().map_or(false, |line| line.is_comment());
+        && lines.peek().is_some_and(|line| line.is_comment());
 
       loop {
         if lines.peek().is_none() {
@@ -299,7 +310,7 @@ impl<'src, D> Recipe<'src, D> {
                 recipe: self.name(),
                 line_number: Some(line_number),
                 code,
-                print_message: self.print_exit_message(),
+                print_message: self.print_exit_message(&context.module.settings),
               });
             }
           } else {
@@ -351,10 +362,8 @@ impl<'src, D> Recipe<'src, D> {
       return Ok(());
     }
 
-    let executor = if let Some(Attribute::Script(interpreter)) = self
-      .attributes
-      .iter()
-      .find(|attribute| matches!(attribute, Attribute::Script(_)))
+    let executor = if let Some(Attribute::Script(interpreter)) =
+      self.attributes.get(AttributeDiscriminant::Script)
     {
       Executor::Command(
         interpreter
@@ -446,7 +455,7 @@ impl<'src, D> Recipe<'src, D> {
               recipe: self.name(),
               line_number: None,
               code,
-              print_message: self.print_exit_message(),
+              print_message: self.print_exit_message(&context.module.settings),
             })
           }
         },
